/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
 * Copyright (C) 2002 Alvaro del Castillo <acs@barrapunto.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <time.h>
#include <glib-object.h>
#include "mrp-intl.h"
#include "mrp-marshal.h"
#include "mrp-storage-module.h"
#include "mrp-task-manager.h"
#include "mrp-private.h"
#include "mrp-time.h"
#include "mrp-error.h"

struct _MrpTaskManagerPriv {
	MrpProject *project;
	MrpTask    *root;

	/* Whether the sorted tree is valid or needs to be rebuilt. */
	gboolean    needs_rebuild;

	/* Whether the task tree needs to be recalculated. */
	gboolean    needs_recalc;
};

typedef struct {
	MrpTaskTraverseFunc func;
	gpointer            user_data;
} MrpTaskTraverseData;

/* Properties */
enum {
	PROP_0,
	PROP_PROJECT,
};

static void mtm_class_init                 (MrpTaskManagerClass *klass);
static void mtm_init                       (MrpTaskManager      *tm);
static void mtm_finalize                   (GObject             *object);
static void mtm_set_property               (GObject             *object,
					    guint                prop_id,
					    const GValue        *value,
					    GParamSpec          *pspec);
static void mtm_get_property               (GObject             *object,
					    guint                prop_id,
					    GValue              *value,
					    GParamSpec          *pspec);
static void mtm_task_duration_changed_cb   (MrpTask             *task,
					    GParamSpec          *spec,
					    MrpTaskManager      *manager);
static void mtm_task_constraint_changed_cb (MrpTask             *task,
					    GParamSpec          *spec,
					    MrpTaskManager      *manager);
static void mtm_project_start_changed_cb   (MrpProject          *project,
					    GParamSpec          *spec,
					    MrpTaskManager      *manager);
static void mtm_task_relation_added_cb     (MrpTask             *task,
					    MrpRelation         *relation,
					    MrpTaskManager      *manager);
static void mtm_task_relation_removed_cb   (MrpTask             *task,
					    MrpRelation         *relation,
					    MrpTaskManager      *manager);
static void mtm_task_relation_changed_cb   (MrpRelation         *relation,
					    GParamSpec          *spec,
					    MrpTaskManager      *manager);
static void mtm_unlink_sorted_tree         (MrpTaskManager      *manager);
static void mtm_dump_task_tree             (GNode               *node);


static GObjectClass *parent_class;


GType
mrp_task_manager_get_type (void)
{
	static GType object_type = 0;

	if (!object_type) {
		static const GTypeInfo object_info = {
			sizeof (MrpTaskManagerClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) mtm_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (MrpTaskManager),
			0,              /* n_preallocs */
			(GInstanceInitFunc) mtm_init,
		};

		object_type = g_type_register_static (G_TYPE_OBJECT,
						      "MrpTaskManager",
						      &object_info,
						      0);
	}

	return object_type;
}

static void
mtm_class_init (MrpTaskManagerClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize     = mtm_finalize;
	object_class->set_property = mtm_set_property;
	object_class->get_property = mtm_get_property;

	/* Properties. */
	g_object_class_install_property (
		object_class,
		PROP_PROJECT,
		g_param_spec_object ("project",
				     _("Project"),
				     _("The project that this task manager is attached to"),
				     G_TYPE_OBJECT,
				     G_PARAM_READWRITE |
				     G_PARAM_CONSTRUCT_ONLY));
}

static void
mtm_init (MrpTaskManager *man)
{
	MrpTaskManagerPriv *priv;
	
	man->priv = g_new0 (MrpTaskManagerPriv, 1);
	priv = man->priv;

	priv->needs_recalc = TRUE;
	priv->needs_rebuild = TRUE;

	priv->root = mrp_task_new ();
}

static void
mtm_finalize (GObject *object)
{
	MrpTaskManager *manager = MRP_TASK_MANAGER (object);

	g_object_unref (manager->priv->root);

	g_free (manager->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize) {
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
	}
}

static void
mtm_set_property (GObject      *object,
		  guint         prop_id,
		  const GValue *value,
		  GParamSpec   *pspec)
{
	MrpTaskManager *manager;
	
	manager = MRP_TASK_MANAGER (object);
	
	switch (prop_id) {
	case PROP_PROJECT:
		/* FIXME: do we need a project here? */
		manager->priv->project = g_value_get_object (value);

		g_signal_connect (manager->priv->project,
				  "notify::project-start",
				  G_CALLBACK (mtm_project_start_changed_cb),
				  manager);
		
		break;

	default:
		break;
	}
}

static void
mtm_get_property (GObject    *object, 
		  guint       prop_id, 
		  GValue     *value,
		  GParamSpec *pspec)
{
	MrpTaskManager *manager;

	manager = MRP_TASK_MANAGER (object);
	
	switch (prop_id) {
	case PROP_PROJECT:
		g_value_set_object (value, manager->priv->project);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

MrpTaskManager *
mrp_task_manager_new (MrpProject *project)
{
	MrpTaskManager *manager;
	
	manager = g_object_new (MRP_TYPE_TASK_MANAGER,
				"project", project,
				NULL);
	
	return manager;
}

static void
mtm_task_connect_signals (MrpTaskManager *manager,
			  MrpTask        *task)
{
	g_signal_connect (task,
			  "notify::duration",
			  G_CALLBACK (mtm_task_duration_changed_cb),
			  manager);
	g_signal_connect (task,
			  "notify::constraint",
			  G_CALLBACK (mtm_task_constraint_changed_cb),
			  manager);
	g_signal_connect (task,
			  "relation-added",
			  G_CALLBACK (mtm_task_relation_added_cb),
			  manager);
	g_signal_connect (task,
			  "relation-removed",
			  G_CALLBACK (mtm_task_relation_removed_cb),
			  manager);	
}

/**
 * mrp_task_manager_insert_task:
 * @manager: A task manager
 * @parent: The parent to insert the task beneath.
 * @position: the position to place task at, with respect to its siblings.
 * If position is -1, task is inserted as the last child of parent.
 * @task: The task to insert.
 * 
 * Inserts a task beneath the parent at the given position.
 * 
 **/
void
mrp_task_manager_insert_task (MrpTaskManager *manager,
			      MrpTask        *parent,
			      gint            position,
			      MrpTask        *task)
{
	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (parent == NULL || MRP_IS_TASK (parent));
	g_return_if_fail (MRP_IS_TASK (task));

	if (parent == NULL) {
		parent = manager->priv->root;
	}

	g_object_set (task,
		      "project", manager->priv->project,
		      NULL);

	imrp_task_insert_child (parent, position, task);

	manager->priv->needs_recalc = TRUE;
	
	imrp_project_task_inserted (manager->priv->project, task);

	mrp_task_manager_recalc (manager);

	mtm_task_connect_signals (manager, task);
}

/**
 * mrp_task_manager_remove_task:
 * @manager: A task manager
 * @task: The task to remove.
 * 
 * Removes a task, or a task subtree if the task has children. The root task
 * (with id 0) cannot be removed.
 *
 **/
void
mrp_task_manager_remove_task (MrpTaskManager *manager,
			      MrpTask        *task)
{
	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (MRP_IS_TASK (task));

	if (task == manager->priv->root) {
		g_warning ("Can't remove root task");
		return;
	}

	g_object_set (task,
		      "project", NULL,
		      NULL);
	
	imrp_task_remove_subtree (task);

	manager->priv->needs_rebuild = TRUE;
	mrp_task_manager_recalc (manager);
}

static gboolean
mtm_get_all_tasks_cb (GNode *node, GList **list)
{
	/* Don't add the root. */
	if (node->parent != NULL) {
		*list = g_list_prepend (*list, node->data);
	}
	
	return FALSE; /* don't stop the traversal */
}

/**
 * mrp_task_manager_get_all_tasks:
 * @manager: A task manager
 * 
 * Gets all the tasks in the project.
 * 
 * Return value: A list of all the MrpTasks in the project.
 **/
GList *
mrp_task_manager_get_all_tasks (MrpTaskManager *manager)
{
	GList   *tasks;

	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), NULL);

	if (manager->priv->root == NULL) {
		return NULL;
	}
	
	tasks = NULL;

	g_node_traverse (imrp_task_get_node (manager->priv->root),
			 G_PRE_ORDER,
			 G_TRAVERSE_ALL,
			 -1,
			 (GNodeTraverseFunc) mtm_get_all_tasks_cb,
			 &tasks);

	tasks = g_list_reverse (tasks);
	
	return tasks;
}

static gboolean
mtm_traverse_cb (GNode *node, MrpTaskTraverseData *data)
{
	return data->func (node->data, data->user_data);
}

/**
 * mrp_task_manager_traverse:
 * @manager: A task manager
 * @root: The task to start traversing
 * @func: A function to call for each traversed task
 * @user_data: Argument to pass to the callback
 * 
 * Calls %func for the subtree starting at %root
 * 
 **/
void
mrp_task_manager_traverse (MrpTaskManager      *manager,
			   MrpTask             *root,
			   MrpTaskTraverseFunc  func,
			   gpointer             user_data)
{
	MrpTaskTraverseData data;

	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (MRP_IS_TASK (root));

	data.func = func;
	data.user_data = user_data;
	
	g_node_traverse (imrp_task_get_node (root),
			 G_PRE_ORDER,
			 G_TRAVERSE_ALL,
			 -1,
			 (GNodeTraverseFunc) mtm_traverse_cb,
			 &data);
}

void
mrp_task_manager_set_root (MrpTaskManager *manager,
			   MrpTask        *task)
{
	GList      *tasks, *l;
	MrpProject *project;
	
	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (task == NULL || MRP_IS_TASK (task));

	if (manager->priv->root != NULL) {
		imrp_task_remove_subtree (manager->priv->root);
	}

	manager->priv->root = task;

	project = manager->priv->project;
	
	tasks = mrp_task_manager_get_all_tasks (manager);
	for (l = tasks; l; l = l->next) {
		g_object_set (l->data, "project", project, NULL);
		
		mtm_task_connect_signals (manager, l->data);
	}

	mrp_task_manager_recalc (manager);
	
	g_list_free (tasks);
}

MrpTask *
mrp_task_manager_get_root (MrpTaskManager *manager)
{
	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), NULL);
	
	return manager->priv->root;
}

static void
mtm_build_tree_copy (GNode *parent_src,
		     GNode *parent_dst,
		     gpointer data)
{
	GNode *child_src;
	GNode *child_dst;

	child_src = g_node_first_child (parent_src);
	while (child_src) {
		child_dst = g_node_new (child_src->data);
		g_node_append (parent_dst, child_dst);

		mtm_build_tree_copy (child_src, child_dst, NULL);
		
		child_src = g_node_next_sibling (child_src);
	}
}

/* Conveniece function for clients that want their own GNode tree. */
GNode *
mrp_task_manager_get_task_tree (MrpTaskManager *manager)
{
	GNode *root_src, *root_dst;
	
	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), NULL);

	root_src = imrp_task_get_node (manager->priv->root);
	root_dst = g_node_new (manager->priv->root);

	mtm_build_tree_copy (root_src, root_dst, NULL);

	/*mtm_dump_task_tree (root_dst);*/
	
	return root_dst;
}

gboolean
mrp_task_manager_move_task (MrpTaskManager  *manager,
			    MrpTask         *task,
			    MrpTask         *sibling,
			    MrpTask         *parent,
			    gboolean         before,
			    GError         **error)
{
	MrpTask *old_parent;
	gint     old_pos;
	MrpTask *grand_parent;
	
	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (task), FALSE);
	g_return_val_if_fail (sibling == NULL || MRP_IS_TASK (sibling), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (parent), FALSE);

	old_parent = mrp_task_get_parent (task);
	old_pos = mrp_task_get_position (task);

	grand_parent = mrp_task_get_parent (old_parent);

	imrp_task_detach (task);

	/* If we change parents, we need to special case a bit:
	 *
	 * 1. We rebuild the sorted tree completely, because it's easier than to
	 *    figure out what to change.
	 *
	 * 2. We must check for loops.
	 *
	 * 2a. Unless, we "unindent", i.e. just change depth one level. That can
	 *     never result in a loop and if we did check it would prevent
	 *     unindenting in some cases.
	 */
	if (parent != old_parent) {
		mtm_unlink_sorted_tree (manager);
		
		if (parent != grand_parent &&
		    !mrp_task_manager_check_move (manager,
						  task,
						  parent,
						  error)) {
			imrp_task_reattach_pos (task, old_pos, old_parent);
			
			mrp_task_manager_rebuild (manager);
			
			return FALSE;
		}
	}

	imrp_task_reattach (task, sibling, parent, before);

	/* If we changed parents, we must rebuild, since we unlinked the sorted
	 * tree above.
	 */
	if (parent != old_parent) {
		mrp_task_manager_rebuild (manager);
	}
	
	imrp_project_task_moved (manager->priv->project,
				 task);
	
	mrp_task_manager_recalc (manager);

	return TRUE;
}

/* ----------------------------------------------------------------------------
 * Test code. Remove at some stage, or move to test framework.
 */

static gchar*
get_n_chars (gint n, gchar c)
{
	GString *str;
	gchar   *ret;
	gint     i;

	str = g_string_new ("");
	
	for (i = 0; i < n; i++) {
		g_string_append_c (str, c);
	}

	ret = str->str;
	g_string_free (str, FALSE);
	
	return ret;
}

static void
dump_children (GNode *node, gint depth)
{
	GNode   *child;
	gchar   *padding = get_n_chars (2 * depth, ' ');
	MrpTask *task;
	gchar   *name;
	
	for (child = g_node_first_child (node); child; child = g_node_next_sibling (child)) {
		task = (MrpTask *) child->data;

		if (MRP_IS_TASK (task)) {
			g_object_get (G_OBJECT (task), "name", &name, NULL);
			g_print ("%sName: %s   ", padding, name);
			g_free (name);

			if (imrp_task_peek_predecessors (task)) {
				GList *l;
				g_print (" <-[");
				for (l = imrp_task_peek_predecessors (task); l; l = l->next) {
					MrpTask *predecessor = mrp_relation_get_predecessor (l->data);

					if (MRP_IS_TASK (predecessor)) {
						g_object_get (G_OBJECT (predecessor), "name", &name, NULL);
						g_print ("%s, ", name);
						g_free (name);
					} else {
						g_print ("<unknown>, ");
					}
				}
				g_print ("]");
			}

			if (imrp_task_peek_successors (task)) {
				GList *l;
				g_print (" ->[");
				for (l = imrp_task_peek_successors (task); l; l = l->next) {
					MrpTask *successor = mrp_relation_get_successor (l->data);

					if (MRP_IS_TASK (successor)) {
						g_object_get (G_OBJECT (successor), "name", &name, NULL);
						g_print ("%s, ", name);
						g_free (name);
					} else {
						g_print ("<unknown>, ");
					}
				}
				g_print ("]");
			}

			g_print ("\n");
		} else {
			g_print ("%s<unknown>\n", padding); 
		}
		
		dump_children (child, depth + 1);
	}

	g_free (padding);
}

static void
mtm_dump_task_tree (GNode *node)
{
	g_return_if_fail (node != NULL);
	g_return_if_fail (node->parent == NULL);
	
	g_print ("------------------------------------------\n<Root>\n");

	dump_children (node, 1);

	g_print ("\n");
}

void
mrp_task_manager_dump_task_tree (MrpTaskManager *manager)
{
	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (manager->priv->root);

	mtm_dump_task_tree (imrp_task_get_node (manager->priv->root));
}

void
mrp_task_manager_dump_task_list (MrpTaskManager *manager)
{
	GList *list, *l;
	gchar *name;

	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (manager->priv->root);

	g_print ("All tasks: ");
	list = mrp_task_manager_get_all_tasks (manager);
	for (l = list; l; l = l->next) {
		if (l != list) {
			g_print (", ");
		}

		if (MRP_IS_TASK (l->data)) {
			g_object_get (G_OBJECT (l->data), "name", &name, NULL);
			g_print ("%s", name);
			g_free (name);
		} else {
			g_print ("<unknown>");
		}
	}
	g_print ("\n");

	g_list_free (list);
}

/* ------------------------------------------------------------------------ */

static gboolean
mtm_unlink_tree_recursive_cb (GNode *node, gpointer data)
{
	g_node_unlink (node);

	return FALSE;
}
	
static void
mtm_unlink_sorted_tree (MrpTaskManager *manager)
{
	GNode *root;

	root = imrp_task_get_sorted_node (manager->priv->root);

	g_node_traverse (root,
			 G_POST_ORDER,
			 G_TRAVERSE_ALL,
			 -1,
			 (GNodeTraverseFunc) mtm_unlink_tree_recursive_cb,
			 NULL);
}

/* Get the ancestor of task_a, that has the same parent as an ancestor or
 * task_b.
 */
static MrpTask *
mtm_get_ancestor_with_same_parent (MrpTask *task_a, MrpTask *task_b)
{
	MrpTask *ancestor;
	gint     depth_a, depth_b, i;

	depth_a = imrp_task_get_depth (task_a);
	depth_b = imrp_task_get_depth (task_b);

	if (depth_a > depth_b) {
		for (i = depth_a; i > depth_b; i--) {
			task_a = mrp_task_get_parent (task_a);
		}
	}
	else if (depth_a < depth_b) {
		for (i = depth_b; i > depth_a; i--) {
			task_b = mrp_task_get_parent (task_b);
		}
	}
	
	ancestor = NULL;
	while (task_a != NULL && task_b != NULL) {

		if (mrp_task_get_parent (task_a) == mrp_task_get_parent (task_b)) {
			ancestor = task_a;
			break;
		}

		task_a = mrp_task_get_parent (task_a);
		task_b = mrp_task_get_parent (task_b);
	}
		
	return ancestor;
}

static void
mtm_traverse_dependency_graph (MrpTaskManager  *manager,
			       MrpTask         *task,
			       GList          **output)
{
	GList       *l;
	MrpRelation *relation;
	MrpTask     *ancestor;
	MrpTask     *child;

	if (imrp_task_get_visited (task)) {
		/*g_warning ("Visited!!\n");*/
		return;
	}

	imrp_task_set_visited (task, TRUE);

	/* Follow successors. */
	for (l = imrp_task_peek_successors (task); l; l = l->next) {
		relation = l->data;

		mtm_traverse_dependency_graph (manager,
					       mrp_relation_get_successor (relation),
					       output);

		/* Also follow the ancestor of the successor that has the same
		 * parent as an ancestor of the task. This way we make sure that
		 * predecessors are listed before the summary tasks of the
		 * successors.
		 */
		ancestor = mtm_get_ancestor_with_same_parent (mrp_relation_get_successor (relation),
							      task);

		if (ancestor != NULL) {
			mtm_traverse_dependency_graph (manager,
						       ancestor,
						       output);
		}
	}
	
	/* Follow parent -> child. */
	child = mrp_task_get_first_child (task);
	while (child) {
		mtm_traverse_dependency_graph (manager, child, output);

		child = mrp_task_get_next_sibling (child);
	}

	if (task != manager->priv->root) {
		/*g_print ("Adding: %s\n", task->priv->name);*/
		*output = g_list_prepend (*output, task);
	}
}

static gboolean
mtm_unset_visited_func (MrpTask  *task,
			gpointer  user_data)
{
	imrp_task_set_visited (task, FALSE);

	return FALSE;
}

static void
mtm_sort_tree (MrpTaskManager *manager)
{
	GList      *list, *l;
	GNode      *root, *node, *parent_node;
	GHashTable *hash;
	MrpTask    *task;

	/* Mark all nodes as unvisited. */
	mrp_task_manager_traverse (manager,
				   manager->priv->root,
				   mtm_unset_visited_func,
				   NULL);
	
	list = NULL;
	mtm_traverse_dependency_graph (manager,
				       manager->priv->root,
				       &list);

	hash = g_hash_table_new (NULL, NULL);

	root = imrp_task_get_sorted_node (manager->priv->root);

	/* Disconnect all the nodes. */
	mtm_unlink_sorted_tree (manager);
	
	g_hash_table_insert (hash, manager->priv->root, root);
	
	for (l = list; l; l = l->next) {
		task = MRP_TASK (l->data);

		node = imrp_task_get_sorted_node (task);

		g_hash_table_insert (hash, task, node);

		parent_node = g_hash_table_lookup (hash,
						   mrp_task_get_parent (task));
		
		g_node_append (parent_node, node);
	}

	g_list_free (list);
	g_hash_table_destroy (hash);

	manager->priv->needs_rebuild = FALSE;
	manager->priv->needs_recalc = TRUE;
	/*dump_task_tree (root);*/
}

/**
 * mtm_get_start:
 * @task: 
 * @project_start: The earliest start to return
 * 
 * Return the start time, calculated from the latest finish of the task's
 * predecessor, or any anscestor's predecessor.
 * 
 * Return value: Calculated start time of task
 **/
static mrptime
mtm_get_start (MrpTask  *task,
	       mrptime   project_start,
	       gboolean *lock_start)
{
	MrpTask       *tmp_task;
	mrptime        finish_plus_lag;
	mrptime        start;
	GList         *predecessors, *l;
	MrpRelation   *relation;
	MrpTask       *predecessor;
	MrpConstraint  constraint;

	start = project_start;

	tmp_task = task;
	while (tmp_task) {
		predecessors = imrp_task_peek_predecessors (tmp_task);
		for (l = predecessors; l; l = l->next) {
			relation = l->data;
			predecessor = mrp_relation_get_predecessor (relation);

			/* Add lag. */
			finish_plus_lag = imrp_task_get_finish (predecessor) +
				mrp_relation_get_lag (relation);
			
			start = MAX (start, finish_plus_lag);
		}
		tmp_task = mrp_task_get_parent (tmp_task);
	}
                     
	constraint = impr_task_get_constraint (task);

	switch (constraint.type) {
	case MRP_CONSTRAINT_SNET: 	/* Start-no-earlier-than. */
		start = MAX (start, constraint.time);
		*lock_start = TRUE;
		break;

	case MRP_CONSTRAINT_MSO:	/* Must-start-on. */
		start = MAX (project_start, constraint.time);
		*lock_start = TRUE;
		break;
		
	case MRP_CONSTRAINT_ASAP:	/* As-soon-as-possible. */
		/* Do nothing. */
		*lock_start = FALSE;
		break;

	case MRP_CONSTRAINT_ALAP:
	case MRP_CONSTRAINT_FNLT:
	default:
		g_warning ("Constraint %d not implemented yet.",
			   constraint.type);
		*lock_start = FALSE;
		break;
	}

	return start;
}

static void
mtm_set_task_start_and_finish (MrpProject *project,
			       MrpTask    *task)
{
	mrptime         t;
	mrptime         t1;
	mrptime         t2;
	mrptime         ival_start;
	mrptime         ival_end;
	GList          *intervals, *l;
	gint             duration;
	gint             work, w, delta;
	MrpCalendar     *calendar;
	MrpCalendarDay  *day;
	MrpInterval     *ival;
	MrpIntervalType  type;
	gboolean         lock_start;
	
	calendar = mrp_project_get_calendar (project);
	
	t1 = mtm_get_start (task,
			    mrp_project_get_project_start (project),
			    &lock_start);

	/* Duration is the work + any nonworking time. */
	work = imrp_task_get_work (task);
	duration = work;
	
	/*g_print ("Work: %.2f\n", work/(60.0*60));*/

	w = 0;
	t = mrp_time_align_day (t1);
	
	while (w < work) {
		day = mrp_calendar_get_day (calendar, t);
		intervals = mrp_calendar_get_intervals (day);
		for (l = intervals; l; l = l->next) { 
			ival = l->data;

			mrp_interval_get_absolute (ival,
						   t,
						   &ival_start,
						   &ival_end);
			
			/* Skip any intervals that are before the task. */
			if (ival_end < t1) {
				continue;
			}

			type = mrp_interval_get_type (ival);
			
			/* If the task starts during nonworking time, delay it
			 * until after the interval (but only if start isn't
			 * locked down by a constraint.
			 */
			if (!lock_start &&
			    type == MRP_INTERVAL_TYPE_NONWORK &&
			    ival_start <= t1 && ival_end > t1) {
				t1 = ival_end;
				
				/* Loop again, in case there is another
				 * nonworking interval right after this one.
				 */
				continue;
			}
				
			/* Clamp the start time so that we don't add work before
			 * the task starts. Note that we don't need to clamp the
			 * end time because it will only be too high during
			 * working time, so the duration won't be affected.
			 */
			delta = ival_end - MAX (ival_start, t1);
			if (type == MRP_INTERVAL_TYPE_WORK) {
				w += delta;
/*				g_print ("[%p] Work ival   : %s - %s, delta: %.2f\n",
				task,
				mrp_time_format ("%a %e %b, %k:%M", ival_start),
				mrp_time_format ("%a %e %b, %k:%M", ival_end),
				delta/(60.0*60));
*/				
			} else {
				duration += delta;
/*				g_print ("[%p] Nonwork ival: %s - %s, delta: %.2f\n",
				task,
				mrp_time_format ("%a %e %b, %k:%M", ival_start),
				mrp_time_format ("%a %e %b, %k:%M", ival_end),
				delta/(60.0*60));
*/			
			}

			if (w >= work) {
				break;
			}
		}

		t += 60*60*24;
	}

	/*duration = imrp_task_get_duration (task);*/

	/*g_print ("Duration: %.2f\n", duration/(60.0*60));*/
	
	imrp_task_set_start (task, t1);
	t2 = t1 + duration;
	imrp_task_set_finish (task, t2);
	imrp_task_set_duration (task, duration);
}

static void
mtm_do_forward_pass (MrpTaskManager *manager,
		     MrpTask        *task,
		     mrptime        *min,
		     mrptime        *max)
{
	MrpTaskManagerPriv *priv;
	GNode              *child;
	mrptime             sub_min, sub_max;
	mrptime             old_start,  old_finish;
	mrptime             tmp_time;

	priv = manager->priv;
	
	/*if (priv->root == task) g_print ("Schedule\n\n");*/
	
	old_start = imrp_task_get_start (task);
	old_finish = imrp_task_get_finish (task);
	
	if (g_node_n_children (imrp_task_get_sorted_node (task)) > 0) {
		/* Summary task. */

		/*g_print ("Summary %s, %d - %d\n", task->priv->name,
		 *min, *max);*/
		
		sub_min = -1;
		sub_max = -1;
		
		child = g_node_first_child (imrp_task_get_sorted_node (task));
		for (; child; child = g_node_next_sibling (child)) {
			mtm_do_forward_pass (manager,
					     child->data,
					     &sub_min,
					     &sub_max);
		}
		
		/* Now the whole subtree has been traversed, set start/finish
		 * from the children.
		 */
		imrp_task_set_start (task, sub_min);
		imrp_task_set_finish (task, sub_max);

		/* FIXME: How should we handle this whole mess with duration vs
		 * start/finish?
		 */
		imrp_task_set_duration (task, sub_max - sub_min);
	} else {
		/* Normal task. */
		mtm_set_task_start_and_finish (priv->project, task);
	}

	/* FIXME: set dirty flag and emit on all changed objects when
	 * we're done?
	 */
	tmp_time = imrp_task_get_start (task);
	if (old_start != tmp_time) {
		g_object_notify (G_OBJECT (task), "start");
	}
	
	tmp_time = imrp_task_get_finish (task);
	if (old_finish != tmp_time) {
		g_object_notify (G_OBJECT (task), "finish");
	}
	
	tmp_time = imrp_task_get_start (task);
	if (*min == -1) {
		*min = tmp_time;
	} else {
		*min = MIN (*min, tmp_time);
	}
	
	tmp_time = imrp_task_get_finish (task);
	if (*max == -1) {
		*max = tmp_time;
	} else {
		*max = MAX (*max, tmp_time);
	}
}

typedef struct {
	GList   *list;
	MrpTask *root;
} GetSortedData;

static gboolean
mtm_get_sorted_tasks_func (GNode         *node,
			   GetSortedData *data)
{
	if (node->data != data->root) {
		data->list = g_list_prepend (data->list, node->data);
	}
	
	return FALSE;
}

static void
mtm_do_backward_pass (MrpTaskManager *manager)
{
	MrpTaskManagerPriv *priv;
	GNode              *root;
	GetSortedData       data;
	GList              *tasks, *l, *s;
	mrptime             project_finish;
	mrptime             t1, t2;
	gint                duration;
	
	priv = manager->priv;

	root = imrp_task_get_sorted_node (priv->root);

	data.list = NULL;
	data.root = root->data;

	g_node_traverse (root,
			 G_POST_ORDER,
			 G_TRAVERSE_ALL,
			 -1,
			 (GNodeTraverseFunc) mtm_get_sorted_tasks_func,
			 &data);

	tasks = data.list;

	project_finish = imrp_task_get_finish (priv->root);
	
	for (l = tasks; l; l = l->next) {
		MrpTask *task = l->data;

		t1 = project_finish;

		for (s = imrp_task_peek_successors (task); s; s = s->next) {
			MrpRelation *relation;
			MrpTask     *successor;

			relation = s->data;
			successor = mrp_relation_get_successor (relation);
			
			t2 = imrp_task_get_latest_start (successor);

			t1 = MIN (t1, t2);
		}

		imrp_task_set_latest_finish (task, t1);

		duration = imrp_task_get_duration (task);
		t1 -= duration; /*mrp_task_duration_to_work (task, duration);? */
		imrp_task_set_latest_start (task, t1);

		t2 = imrp_task_get_start (task);
		
		g_object_set (task, "critical", t1 == t2, NULL);
	}

	g_list_free (tasks);
}

void
mrp_task_manager_rebuild (MrpTaskManager *manager)
{
	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (manager->priv->root != NULL);

	mtm_sort_tree (manager);
	manager->priv->needs_rebuild = FALSE;
	manager->priv->needs_recalc = TRUE;
}

void
mrp_task_manager_recalc (MrpTaskManager *manager)
{
	glong min, max;

	g_return_if_fail (MRP_IS_TASK_MANAGER (manager));
	g_return_if_fail (manager->priv->root != NULL);

	if (!manager->priv->needs_recalc && !manager->priv->needs_rebuild) {
		return;
	}
	
	if (manager->priv->needs_rebuild) {
		mrp_task_manager_rebuild (manager);
	}
	
	min = G_MAXLONG;
	max = 0;

	mtm_do_forward_pass (manager, manager->priv->root, &min, &max);
	mtm_do_backward_pass (manager);
	manager->priv->needs_recalc = FALSE;
}

static void
mtm_task_duration_changed_cb (MrpTask        *task,
			      GParamSpec     *spec,
			      MrpTaskManager *manager)
{
	manager->priv->needs_recalc = TRUE;
	mrp_task_manager_recalc (manager);
}

static void
mtm_task_constraint_changed_cb (MrpTask        *task,
				GParamSpec     *spec,
				MrpTaskManager *manager)
{
	manager->priv->needs_recalc = TRUE;
	mrp_task_manager_recalc (manager);
}

static void
mtm_project_start_changed_cb (MrpProject     *project,
			      GParamSpec     *spec,
			      MrpTaskManager *manager)
{
	manager->priv->needs_recalc = TRUE;
	mrp_task_manager_recalc (manager);
}

static void
mtm_task_relation_changed_cb (MrpRelation    *relation,
			      GParamSpec     *spec,
			      MrpTaskManager *manager)
{
	MrpTask *task;
	
	task = mrp_relation_get_successor (relation);
	manager->priv->needs_recalc = TRUE;
	
	mrp_task_manager_recalc (manager);	
}



static void
mtm_task_relation_added_cb (MrpTask        *task,
			    MrpRelation    *relation,
			    MrpTaskManager *manager)
{
	/* We get a signal on both the predecessor and the successor, it's
	 * enough with one rebuild.
	 */
	if (task == mrp_relation_get_predecessor (relation)) {
		return;
	}
	g_signal_connect_object (relation,
				 "notify",
				 G_CALLBACK (mtm_task_relation_changed_cb),
				 manager, 0);

	manager->priv->needs_rebuild = TRUE;
	mrp_task_manager_recalc (manager);
}

static void
mtm_task_relation_removed_cb (MrpTask        *task,
			      MrpRelation    *relation,
			      MrpTaskManager *manager)
{
	/* We get a signal on both the predecessor and the successor, it's
	 * enough with one rebuild.
	 */
	if (task == mrp_relation_get_predecessor (relation)) {
		return;
	}
	
	g_signal_handlers_disconnect_by_func (MRP_OBJECT(relation),
					      mtm_task_relation_changed_cb,
					      manager);
	
	manager->priv->needs_rebuild = TRUE;
	mrp_task_manager_recalc (manager);
}

static gboolean
mtm_check_successor_loop (MrpTask *task, MrpTask *end_task)
{
	GList       *successors, *l;
	MrpRelation *relation;
	MrpTask     *child;

	if (task == end_task) {
		return FALSE;
	}
	
	successors = imrp_task_peek_successors (task);
	for (l = successors; l; l = l->next) {
		relation = l->data;
		if (!mtm_check_successor_loop (mrp_relation_get_successor (relation),
					       end_task)) {
			return FALSE;
		}
	}

	child = mrp_task_get_first_child (task);
	while (child) {
		if (!mtm_check_successor_loop (child, end_task)) {
			return FALSE;
		}
		
		child = mrp_task_get_next_sibling (child);
	}
	
	return TRUE;
}

static gboolean
mtm_check_predecessor_loop (MrpTask *task, MrpTask *end_task)
{
	GList       *predecessors, *l;
	MrpRelation *relation;
	MrpTask     *child;

	if (task == end_task) {
		return FALSE;
	}
	
	predecessors = imrp_task_peek_predecessors (task);
	for (l = predecessors; l; l = l->next) {
		relation = l->data;
		if (!mtm_check_predecessor_loop (mrp_relation_get_predecessor (relation),
						 end_task)) {
			return FALSE;
		}
	}

	child = mrp_task_get_first_child (task);
	while (child) {
		if (!mtm_check_predecessor_loop (child, end_task)) {
			return FALSE;
		}
		
		child = mrp_task_get_next_sibling (child);
	}
	
	return TRUE;
}

gboolean
mrp_task_manager_check_predecessor (MrpTaskManager  *manager,
				    MrpTask         *task,
				    MrpTask         *predecessor,
				    GError         **error)
{
	gint     depth_task, depth_predecessor;
	gint     i;
	MrpTask *task_ancestor;
	MrpTask *predecessor_ancestor;

	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (task), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (predecessor), FALSE);

	/* "Inbreed" check, i.e. a relation between a task and it's ancestor. */
	depth_task = imrp_task_get_depth (task);
	depth_predecessor = imrp_task_get_depth (predecessor);

	task_ancestor = task;
	predecessor_ancestor = predecessor;

	if (depth_task < depth_predecessor) {
		for (i = depth_predecessor; i > depth_task; i--) {
			predecessor_ancestor = mrp_task_get_parent (predecessor_ancestor);
		}
	} else if (depth_predecessor < depth_task) {
		for (i = depth_task; i > depth_predecessor; i--) {
			task_ancestor = mrp_task_get_parent (task_ancestor);
		}
	}

	if (task_ancestor == predecessor_ancestor) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_TASK_RELATION_FAILED,
			     _("Can not add a predecessor relation between a task and its ancestor."));
		return FALSE;
	}

	/* Loop checking. */
	if (!mtm_check_successor_loop (task, predecessor)) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_TASK_RELATION_FAILED,
			     _("Can not add a predecessor, because it would result in a loop."));
		return FALSE;
	}

	if (!mtm_check_predecessor_loop (predecessor, task)) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_TASK_RELATION_FAILED,
			     _("Can not add a predecessor, because it would result in a loop."));
		return FALSE;
	}

	return TRUE;
}

gboolean
mrp_task_manager_check_move (MrpTaskManager  *manager,
			     MrpTask         *task,
			     MrpTask         *parent,
			     GError         **error)
{
	gboolean success;
	
	g_return_val_if_fail (MRP_IS_TASK_MANAGER (manager), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (task), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (parent), FALSE);

	/* If there is a link between the task's subtree and the new parent or
	 * any of its ancestors, the move will result in a loop.
	 */
	success = mrp_task_manager_check_predecessor (manager,
						      task,
						      parent,
						      NULL) &&
		mrp_task_manager_check_predecessor (manager,
						    parent,
						    task,
						    NULL);

	if (!success) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_TASK_MOVE_FAILED,
			     _("Can not move the task, since it would result in a loop."));
		return FALSE;
	}

	return TRUE;
}


