/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <string.h>
#include "mrp-marshal.h"
#include "mrp-intl.h"
#include "mrp-private.h"
#include "mrp-resource.h"
#include "mrp-error.h"

/* Properties */
enum {
	PROP_0,
	PROP_NAME,
	PROP_START,
	PROP_FINISH,
	PROP_LATEST_START,
	PROP_LATEST_FINISH,
	PROP_DURATION,
	PROP_WORK,
	PROP_CRITICAL,
	PROP_CONSTRAINT,
	PROP_NOTE,
	PROP_PERCENT_COMPLETE
};

/* Signals */
enum {
	RELATION_ADDED,
	RELATION_REMOVED,
	TASK_MOVED,
	ASSIGNMENT_ADDED,
	ASSIGNMENT_REMOVED,
	CHILD_ADDED,
	CHILD_REMOVED,
	LAST_SIGNAL
};

struct _MrpTaskPriv {
	MrpTaskManager   *manager;

	gchar            *name;
	gchar            *note;

	/* Percent complete, 0-100. */
	gshort            percent_complete;
	
	/* The amount of work effort (duration*units) necessary to complete this
	 * task.
	 */
	gint              work;

	/* Calculated duration. */
	gint              duration;

	/* Calculated start and finish values. */
	mrptime           start;
	mrptime           finish;

	gboolean          critical;
	gint              slack;
	
	/* Represents the Work Breakdown Structure tree hierarchy. */
	GNode            *node;

	/* Used for topological order sorting. */
	GNode            *sorted_node;
	gboolean          visited;

	/* The acyclic dependency graph. */
	GList            *successors;
	GList            *predecessors;
	
	/* Calculated CPM values. */
	mrptime           latest_start;
	mrptime           latest_finish;

	MrpConstraint     constraint;

	/* List of assignments. */
	GHashTable       *assignments;
};


static void mpt_init                  (MrpTask            *module);
static void mpt_class_init            (MrpTaskClass       *class);
static void mpt_finalize              (GObject            *object);
static void mpt_set_property          (GObject            *object,
				       guint               prop_id,
				       const GValue       *value,
				       GParamSpec         *pspec);
static void mpt_get_property          (GObject            *object,
				       guint               prop_id,
				       GValue             *value,
				       GParamSpec         *pspec);
static void mpt_removed               (MrpObject          *object);
static void mpt_assignment_removed_cb (MrpAssignment      *assignment,
				       MrpTask            *task);


static MrpObjectClass *parent_class;
static guint signals[LAST_SIGNAL];


GType
mrp_task_get_type (void)
{
	static GType object_type = 0;

	if (!object_type) {
		static const GTypeInfo object_info = {
			sizeof (MrpTaskClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) mpt_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (MrpTask),
			0,              /* n_preallocs */
			(GInstanceInitFunc) mpt_init,
		};

		object_type = g_type_register_static (MRP_TYPE_OBJECT,
						      "MrpTask",
						      &object_info, 0);
	}

	return object_type;
}

static void
mpt_init (MrpTask *task)
{
	task->priv = g_new0 (MrpTaskPriv, 1);

	task->priv->name = g_strdup ("");
	task->priv->node = g_node_new (task);
	task->priv->sorted_node = g_node_new (task);
	task->priv->assignments = g_hash_table_new (NULL, NULL);
	task->priv->constraint.type = MRP_CONSTRAINT_ASAP;
}

static void
mpt_class_init (MrpTaskClass *klass)
{
	GObjectClass   *object_class     = G_OBJECT_CLASS (klass);
	MrpObjectClass *mrp_object_class = MRP_OBJECT_CLASS (klass);
	
	parent_class = MRP_OBJECT_CLASS (g_type_class_peek_parent (klass));

	object_class->finalize = mpt_finalize;
	object_class->set_property = mpt_set_property;
	object_class->get_property = mpt_get_property;

	mrp_object_class->removed  = mpt_removed;
	
	signals[TASK_MOVED] =
		g_signal_new ("task-moved",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__OBJECT_INT,
			      G_TYPE_NONE, 2, MRP_TYPE_TASK, G_TYPE_INT);
	
	signals[RELATION_ADDED] =
		g_signal_new ("relation-added",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__POINTER,
			      G_TYPE_NONE, 1, G_TYPE_POINTER);
	
	signals[RELATION_REMOVED] =
		g_signal_new ("relation-removed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__POINTER,
			      G_TYPE_NONE, 1, G_TYPE_POINTER);

	signals[ASSIGNMENT_ADDED] = 
		g_signal_new ("assignment-added",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__OBJECT,
			      G_TYPE_NONE,
			      1, MRP_TYPE_ASSIGNMENT);

	signals[ASSIGNMENT_REMOVED] = 
		g_signal_new ("assignment-removed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__OBJECT,
			      G_TYPE_NONE,
			      1, MRP_TYPE_ASSIGNMENT);
	
	signals[CHILD_ADDED] = 
		g_signal_new ("child-added",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	signals[CHILD_REMOVED] = 
		g_signal_new ("child-removed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      mrp_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	g_object_class_install_property (
		object_class,
		PROP_NAME,
		g_param_spec_string ("name",
				     _("Task name"),
				     NULL,
				     "",
				     G_PARAM_READWRITE));
	
	g_object_class_install_property (
		object_class,
		PROP_NOTE,
		g_param_spec_string ("note",
				     _("Task note"),
				     NULL,
				     "",
				     G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_START,
		mrp_param_spec_time ("start",
				     _("Task start time"),
				     NULL,
				     G_PARAM_READABLE));
	
	g_object_class_install_property (
		object_class,
		PROP_FINISH,
		mrp_param_spec_time ("finish",
				     _("Task finish time"),
				     NULL,
				     G_PARAM_READABLE));
	
	g_object_class_install_property (
		object_class,
		PROP_LATEST_START,
		mrp_param_spec_time ("latest-start",
				     _("Latest task start time"),
				     NULL,
				     G_PARAM_READABLE));
	
	g_object_class_install_property (
		object_class,
		PROP_LATEST_FINISH,
		mrp_param_spec_time ("latest-finish",
				     _("Latest task finish time"),
				     NULL,
				     G_PARAM_READABLE));
	
	g_object_class_install_property (
		object_class,
		PROP_DURATION,
		g_param_spec_int ("duration",
				  _("Task duration"),
				  NULL,
				  -1, G_MAXINT, 0,
				  G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_WORK,
		g_param_spec_int ("work",
				  _("Task work"),
				  NULL,
				  -1, G_MAXINT, 0,
				  G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_CRITICAL,
		g_param_spec_boolean ("critical",
				      _("In critical path"),
				      NULL,
				      FALSE,
				      G_PARAM_READWRITE));

	g_object_class_install_property (
		object_class,
		PROP_CONSTRAINT,
		g_param_spec_boxed ("constraint",
				    _("Time constraint"),
				    NULL,
				    MRP_TYPE_CONSTRAINT,
				    G_PARAM_READWRITE));
	
	g_object_class_install_property (
		object_class,
		PROP_PERCENT_COMPLETE,
		g_param_spec_int ("percent-complete",
				  _("Percent complete"),
				  NULL,
				  0, 100, 0,
				  G_PARAM_READWRITE));
}

static void
mpt_set_property (GObject      *object,
		  guint         prop_id,
		  const GValue *value,
		  GParamSpec   *pspec)
{
	MrpTask     *task;
	MrpTaskPriv *priv;
	const gchar *str;
	gint         i_val;
	gboolean     changed = FALSE;
	gint         units;

	task = MRP_TASK (object);
	priv = task->priv;
	
	switch (prop_id) {
	case PROP_NAME:
		str = g_value_get_string (value);

		if (!priv->name || strcmp (priv->name, str)) {
			g_free (task->priv->name);
			priv->name = g_strdup (str);
			changed = TRUE;
		}
		
		break;

	case PROP_NOTE:
		str = g_value_get_string (value);

		if (!priv->note || strcmp (priv->note, str)) {
			g_free (task->priv->note);
			priv->note = g_strdup (str);
			changed = TRUE;
		}

		break;
		
/* FIXME: These are all fake values for now, fix when the task
 * math code is implemented.
 */
		
 	case PROP_START:
		priv->start = g_value_get_long (value);
		break;

	case PROP_FINISH:
		priv->finish = g_value_get_long (value);
		break;

	case PROP_DURATION:
		if (mrp_task_get_n_children (task) > 0) {
			g_warning ("Can't set duration on a summary task.");
			return;
		}

		i_val = g_value_get_int (value);
		
		if (priv->duration != i_val) {
			priv->duration = i_val;

			units = 1;
			priv->work = priv->duration * units;
			g_object_notify (object, "work");
			
			changed = TRUE;
		}
		break;
		
	case PROP_WORK:
		if (mrp_task_get_n_children (task) > 0) {
			g_warning ("Can't set work on a summary task.");
			return;
		}
		
		i_val = g_value_get_int (value);
		
		if (priv->work != i_val) {
			priv->work = i_val;

			units = 1;
			priv->duration = priv->work / units;
			g_object_notify (object, "duration");

			changed = TRUE;
		}
		break;

/* End of fake.
 */
		
	case PROP_CRITICAL:
		priv->critical = g_value_get_boolean (value);
		break;

 	case PROP_CONSTRAINT:
		/* FIXME: compare */
		priv->constraint = *(MrpConstraint *) g_value_get_boxed (value);
		changed = TRUE;
		break;

	case PROP_PERCENT_COMPLETE:
		i_val = g_value_get_int (value);
		
		if (priv->percent_complete != i_val) {
			priv->percent_complete = i_val;
			changed = TRUE;
		}
		
		break;

	default:
		break;
	}

	if (changed) {
		mrp_object_set_needs_saving (MRP_OBJECT (object), TRUE);
	}
}

static void
mpt_get_property (GObject    *object,
		  guint       prop_id,
		  GValue     *value,
		  GParamSpec *pspec)
{
	MrpTask     *task;
	MrpTaskPriv *priv;

	task = MRP_TASK (object);
	priv = task->priv;
	
	switch (prop_id) {
	case PROP_NAME:
		g_value_set_string (value, priv->name);
		break;
	case PROP_NOTE:
		g_value_set_string (value, priv->note);
		break;
	case PROP_START:
		g_value_set_long (value, priv->start);
		break;
	case PROP_FINISH:
		g_value_set_long (value, priv->finish);
		break;
	case PROP_LATEST_START:
		g_value_set_long (value, priv->latest_start);
		break;
	case PROP_LATEST_FINISH:
		g_value_set_long (value, priv->latest_finish);
		break;
	case PROP_DURATION:
		g_value_set_int (value, priv->duration);
		break;
	case PROP_WORK:
		g_value_set_int (value, priv->work);
		break;
	case PROP_CRITICAL:
		g_value_set_boolean (value, priv->critical);
		break;
	case PROP_CONSTRAINT:
		g_value_set_boxed (value, &priv->constraint);
		break;
	case PROP_PERCENT_COMPLETE:
		g_value_set_int (value, priv->percent_complete);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
mpt_remove_assignment_foreach (gpointer key,
			       gpointer value, 
			       gpointer user_data)
{
	g_return_if_fail (MRP_IS_ASSIGNMENT (value));
	
	g_signal_handlers_disconnect_by_func (MRP_OBJECT (value),
					      mpt_assignment_removed_cb,
					      MRP_TASK (user_data));
	mrp_object_removed (MRP_OBJECT (value));
}

static void
mpt_removed (MrpObject *object)
{
	MrpTask     *task;
	MrpTaskPriv *priv;
	
	g_return_if_fail (MRP_IS_TASK (object));
	
	task = MRP_TASK (object);
	priv = task->priv;
	
	g_hash_table_foreach (priv->assignments, 
			      (GHFunc) mpt_remove_assignment_foreach,
			      task);
	
        if (MRP_OBJECT_CLASS (parent_class)->removed) {
                (* MRP_OBJECT_CLASS (parent_class)->removed) (object);
        }
}

static void
mpt_assignment_removed_cb (MrpAssignment *assignment, MrpTask *task)
{
	MrpTaskPriv *priv;
	MrpResource *resource;
	
	g_return_if_fail (MRP_IS_TASK (task));
	g_return_if_fail (MRP_IS_ASSIGNMENT (assignment));
	
	priv = task->priv;

	resource = mrp_assignment_get_resource (assignment);
	
	if (!resource) {
		g_warning ("Resource not found in tasks assignments list");
		return;
	}
	
	g_hash_table_remove (priv->assignments, resource);
	g_signal_emit (task, signals[ASSIGNMENT_REMOVED], 0, assignment);
	g_object_unref (assignment);

	mrp_object_set_needs_saving (MRP_OBJECT (task), TRUE);
}

static void
mpt_finalize (GObject *object)
{
	MrpTask     *task;
	MrpTaskPriv *priv;

	task = MRP_TASK (object);
	priv = task->priv;
	
	g_free (priv->name);
	g_free (priv->note);

	/* Make sure we aren't left hanging in the tree. */
	g_assert (priv->node->parent == NULL);

	/* Make sure we don't have dangling relations. */
	g_assert (priv->predecessors == NULL);
	g_assert (priv->successors == NULL);
	
	g_node_destroy (priv->node);
	g_node_destroy (priv->sorted_node);

	g_free (priv);
	task->priv = NULL;

	if (G_OBJECT_CLASS (parent_class)->finalize) {
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
	}
}

static void
mpt_remove_relations (MrpTask *task)
{
	GList       *l, *next;
	MrpRelation *relation;
	MrpTask     *predecessor;
	MrpTask     *successor;

	g_return_if_fail (MRP_IS_TASK (task));
	
	/* Cut relations involving the task (make sure to be robust when
	 * removing during traversing).
	 */
	l = task->priv->predecessors;
	while (l) {
		next = l->next;
		relation = l->data;

		predecessor = mrp_relation_get_predecessor (relation);
		
		mrp_task_remove_predecessor (task, predecessor);

		l = next;
	}
	
	l = task->priv->successors;
	while (l) {
		next = l->next;
		relation = l->data;

		successor = mrp_relation_get_successor (relation);
		
		mrp_task_remove_predecessor (successor, task);

		l = next;
	}
}

static gboolean
mpt_remove_subtree_cb (GNode *node, gpointer data)
{
	MrpTask     *task;
	MrpTaskPriv *priv;

	task = node->data;
	priv = task->priv;
	
	mpt_remove_relations (task);

	g_node_unlink (priv->node);
	g_node_unlink (priv->sorted_node);

	g_object_unref (task);
	
	return FALSE;
}
				 
void
imrp_task_remove_subtree (MrpTask *task)
{
	MrpTask *parent;
	
	g_return_if_fail (MRP_IS_TASK (task));

	parent = NULL;
	if (task->priv->node->parent) {
		parent = task->priv->node->parent->data;
	}
	
	g_object_ref (task);
	
	/* Remove the tasks one by one using post order so we don't mess with
	 * the tree while traversing it.
	 */
	g_node_traverse (task->priv->node,
			 G_POST_ORDER,
			 G_TRAVERSE_ALL,
			 -1,
			 (GNodeTraverseFunc) mpt_remove_subtree_cb,
			 NULL);

	g_object_unref (task);

	if (parent) {
		g_signal_emit (parent, signals[CHILD_REMOVED], 0);
	}
}

void
imrp_task_detach (MrpTask *task)
{
	g_return_if_fail (MRP_IS_TASK (task));

	/* FIXME: Do some extra checking. */

	g_node_unlink (task->priv->node);

	/* Note: we don't unlink the sorted node, because when moving a task but
	 * keeping it under the same parent, we don't need to rebuild the
	 * tree. So when the caller wants that, it has to unlink the tree and
	 * rebuild it.
	 */
}

void
imrp_task_reattach (MrpTask  *task,
		    MrpTask  *sibling,
		    MrpTask  *parent,
		    gboolean  before)
{
	GNode *node;
	
	g_return_if_fail (MRP_IS_TASK (task));
	g_return_if_fail (sibling == NULL || MRP_IS_TASK (sibling));
	g_return_if_fail (MRP_IS_TASK (parent));

	/* FIXME: Do some extra checking. */

	if (sibling == NULL) {
		if (before) {
			node = g_node_first_child (parent->priv->node);
		} else {
			node = g_node_last_child (parent->priv->node);
		}			

		if (node) {
			sibling = node->data;
		}
	}

	if (before) {
		if (sibling != NULL) {
			g_node_insert_before (parent->priv->node,
					      sibling->priv->node,
					      task->priv->node);
		} else {
			g_node_prepend (parent->priv->node,
					task->priv->node);
		}
	} else {
		if (sibling != NULL) {
			g_node_insert_after (parent->priv->node,
					     sibling->priv->node,
					     task->priv->node);
		} else {
			g_node_append (parent->priv->node,
				       task->priv->node);
		}
	}
}

void
imrp_task_reattach_pos (MrpTask  *task,
			gint      pos,
			MrpTask  *parent)
{
	g_return_if_fail (MRP_IS_TASK (task));
	g_return_if_fail (pos >= -1);
	g_return_if_fail (MRP_IS_TASK (parent));

	/* FIXME: Do some extra checking. */

	g_node_insert (parent->priv->node,
		       pos,
		       task->priv->node);
}

MrpTask *
mrp_task_new (void)
{
	MrpTask *task;

	task = g_object_new (MRP_TYPE_TASK, NULL);

	return task;
}

void
imrp_task_insert_child (MrpTask *parent,
			gint     position,
			MrpTask *child)
{
	g_return_if_fail (MRP_IS_TASK (parent));
	g_return_if_fail (MRP_IS_TASK (child));

	if (child->priv->duration == -1) {
		child->priv->duration = parent->priv->duration;
	}
	
	g_node_insert (parent->priv->node,
		       position,
		       child->priv->node);

	g_node_prepend (parent->priv->sorted_node,
			child->priv->sorted_node);

	g_signal_emit (parent, signals[CHILD_ADDED], 0);
}

static MrpRelation *
mpt_get_predecessor_relation (MrpTask *task,
			      MrpTask *predecessor)
{
	GList *l;

	for (l = task->priv->predecessors; l; l = l->next) {
		MrpRelation *relation = l->data;

		if (mrp_relation_get_successor (relation) == task &&
		    mrp_relation_get_predecessor (relation) == predecessor) {
			return relation;
		}
	}

	return NULL;
}

static MrpRelation *
mpt_get_successor_relation (MrpTask *task,
			    MrpTask *successor)
{
	GList *l;

	for (l = task->priv->successors; l; l = l->next) {
		MrpRelation *relation = l->data;

		if (mrp_relation_get_predecessor (relation) == task &&
		    mrp_relation_get_successor (relation) == successor) {
			return relation;
		}
	}

	return NULL;
}

gboolean
mrp_task_has_relation_to (MrpTask *a, MrpTask *b)
{
	return (mpt_get_predecessor_relation (a, b) != NULL) ||
		(mpt_get_predecessor_relation (b, a) != NULL);
}

gboolean
mrp_task_has_relation (MrpTask *task)
{
	return (task->priv->predecessors != NULL ||
		task->priv->successors != NULL);
}

/**
 * mrp_task_add_predecessor:
 * @task: The task to add predecessor to
 * @predecessor: The predecessor
 * @type: Type of relation
 * @lag: Lag time. If negative, it means lead time.
 * 
 * Adds a predecessor task to a task. Depending on type, the predecessor
 * must be started or finished before task can be started or finished,
 * with an optional lag/lead.
 *
 * Return value: The relation that represents the predecessor/successor link.
 **/
MrpRelation *
mrp_task_add_predecessor (MrpTask          *task,
			  MrpTask          *predecessor,
			  MrpRelationType   type,
			  glong             lag,
			  GError          **error)
{
	MrpRelation    *relation;
	MrpProject     *project;
	MrpTaskManager *manager;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	g_return_val_if_fail (MRP_IS_TASK (predecessor), NULL);

	/* Can't have more than one relation between the same two tasks. */
	if (mrp_task_has_relation_to (task, predecessor)) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_TASK_RELATION_FAILED,
			     "Could not add a predecessor relation, because the tasks are already related.");
		
		return NULL;
	}

	g_object_get (G_OBJECT (task), "project", &project, NULL);
	manager = mrp_project_get_task_manager (project);
	if (!mrp_task_manager_check_predecessor (manager, task, predecessor, error)) {
		return NULL;
	}
	
	relation = mrp_relation_new (task, predecessor, type, lag);

	task->priv->predecessors = g_list_prepend (task->priv->predecessors, relation);
	predecessor->priv->successors = g_list_prepend (predecessor->priv->successors, relation);

	g_signal_emit (task, signals[RELATION_ADDED], 0, relation);
	g_signal_emit (predecessor, signals[RELATION_ADDED], 0, relation);

	mrp_object_set_needs_saving (MRP_OBJECT (task), TRUE);
	mrp_object_set_needs_saving (MRP_OBJECT (predecessor), TRUE);

	return relation;
}

/**
 * mrp_task_remove_predecessor:
 * @task: The task to remove a predecessor from
 * @predecessor: The predecessor to remove
 * 
 * Removes a predecessor previously added to task.
 **/
void
mrp_task_remove_predecessor (MrpTask *task,
			     MrpTask *predecessor)
{
	MrpRelation *relation;
	
	g_return_if_fail (MRP_IS_TASK (task));
	g_return_if_fail (MRP_IS_TASK (predecessor));

	relation = mpt_get_predecessor_relation (task, predecessor);

	task->priv->predecessors = g_list_remove (task->priv->predecessors, relation);
	predecessor->priv->successors = g_list_remove (predecessor->priv->successors, relation);

	/* Notify. */
	mrp_object_removed (MRP_OBJECT (relation));

	g_signal_emit (task, signals[RELATION_REMOVED], 0, relation);
	g_signal_emit (predecessor, signals[RELATION_REMOVED], 0, relation);

	mrp_object_set_needs_saving (MRP_OBJECT (task), TRUE);
	mrp_object_set_needs_saving (MRP_OBJECT (predecessor), TRUE);

	/* FIXME: ? */
	g_object_unref (relation);
}

/**
 * mrp_task_get_relation:
 * @task_a: A task 
 * @task_b: A task
 * 
 * 
 * Return value: Any relation between two tasks.
 **/
MrpRelation *
mrp_task_get_relation (MrpTask *task_a,
		       MrpTask *task_b)
{
	MrpRelation *relation;
	
	g_return_val_if_fail (MRP_IS_TASK (task_a), NULL);
	g_return_val_if_fail (MRP_IS_TASK (task_b), NULL);
	
	relation = mpt_get_predecessor_relation (task_a, task_b);
	if (relation != NULL) {
		return relation;
	}

	return mpt_get_successor_relation (task_a, task_b);
}

/**
 * mrp_task_get_predecessor:
 * @task: 
 * @predecessor: 
 * 
 * 
 * 
 * Return value: 
 **/
MrpRelation *
mrp_task_get_predecessor (MrpTask *task,
			  MrpTask *predecessor)
{
	MrpRelation *relation;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	g_return_val_if_fail (MRP_IS_TASK (predecessor), NULL);
	
	relation = mpt_get_predecessor_relation (task, predecessor);
	return relation;
}

/**
 * mrp_task_get_successor:
 * @task: 
 * @successor: 
 * 
 * 
 * 
 * Return value: 
 **/
MrpRelation *
mrp_task_get_successor (MrpTask *task,
			MrpTask *successor)
{
	MrpRelation *relation;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	g_return_val_if_fail (MRP_IS_TASK (successor), NULL);
	
	relation = mpt_get_successor_relation (task, successor);
	return relation;
}

GList *
mrp_task_get_predecessors (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	return g_list_copy (task->priv->predecessors);
}

GList *
mrp_task_get_successors (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	return g_list_copy (task->priv->successors);
}

MrpTask *
mrp_task_get_parent (MrpTask *task)
{
	GNode *node;

	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	node = task->priv->node->parent;
	return node ? node->data : NULL;
}

MrpTask *
mrp_task_get_first_child (MrpTask *task)
{
	GNode *node;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	node = g_node_first_child (task->priv->node);
	return node ? node->data : NULL;
}

MrpTask *
mrp_task_get_next_sibling (MrpTask *task)
{
	GNode *node;

	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	node = g_node_next_sibling (task->priv->node);
	return node ? node->data : NULL;
}

guint
mrp_task_get_n_children (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), 0);

	return g_node_n_children (task->priv->node);
}

MrpTask *
mrp_task_get_nth_child (MrpTask *task, guint n)
{
	GNode *node;

	g_return_val_if_fail (MRP_IS_TASK (task), 0);

	node = g_node_nth_child (task->priv->node, n);
	return node ? node->data : NULL;
}

gint
mrp_task_get_position (MrpTask *task)
{
	GNode *parent;
	
	g_return_val_if_fail (MRP_IS_TASK (task), 0);
	g_return_val_if_fail (task->priv->node->parent != NULL, 0);

	parent = task->priv->node->parent;
	
	return g_node_child_position (parent, task->priv->node);
}

void
mrp_task_add_assignment (MrpTask       *task,
			 MrpAssignment *assignment)
{
	MrpTaskPriv *priv;
	
	g_return_if_fail (MRP_IS_TASK (task));
	g_return_if_fail (MRP_IS_ASSIGNMENT (assignment));
	
	priv = task->priv;
	
	g_hash_table_insert (priv->assignments, 
			     mrp_assignment_get_resource (assignment),
			     g_object_ref (assignment));

	g_signal_connect (assignment,
			  "removed",
			  G_CALLBACK (mpt_assignment_removed_cb),
			  task);

	g_signal_emit (task, signals[ASSIGNMENT_ADDED], 0, assignment);

	mrp_object_set_needs_saving (MRP_OBJECT (task), TRUE);
}

MrpTimeInterval *
mrp_task_get_interval (MrpTask *task)
{
	/* FIXME: Implement */
	return NULL;
}

mrptime
mrp_task_get_start (MrpTask *task)
{
	/*g_return_val_if_fail (MRP_IS_TASK (task), NULL);*/

	/* FIXME: this will be a calculated value. */
	return task->priv->start;
}

mrptime
mrp_task_get_finish (MrpTask *task)
{
	/*g_return_val_if_fail (MRP_IS_TASK (task), NULL);*/

	/* FIXME: this will be a calculated value. */
	return task->priv->finish;
}

mrptime
imrp_task_get_latest_start (MrpTask *task)
{
	return task->priv->latest_start;
}

mrptime
imrp_task_get_latest_finish (MrpTask *task)
{
	return task->priv->latest_finish;
}

void
imrp_task_set_latest_start (MrpTask *task,
			    mrptime  time)
{
	task->priv->latest_start = time;
}

void
imrp_task_set_latest_finish (MrpTask *task,
			     mrptime  time)
{
	task->priv->latest_finish = time;
}

gint
mrp_task_get_duration (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), -1);

	if (g_node_n_children (task->priv->node) == 0) {
		/* Leaf, duration is what we say it is. */
		return task->priv->duration;
	}
	
	/* Summary, we need a calculated value. */
	return task->priv->duration;
}

static void
foreach_assignment_add_to_list (gpointer key,
				gpointer value, 
				gpointer user_data)
{
	GList **list;
	
	list = (GList **) user_data;
	
	*list = g_list_prepend (*list, value);
}

GList *
mrp_task_get_assignments (MrpTask *task)
{
	GList *list = NULL;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	
	g_hash_table_foreach (task->priv->assignments, 
			      foreach_assignment_add_to_list,
			      &list);
			      
	return list;
}

MrpAssignment *
mrp_task_get_assignment (MrpTask *task, MrpResource *resource)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	g_return_val_if_fail (MRP_IS_RESOURCE (resource), NULL);

	return g_hash_table_lookup (task->priv->assignments, resource);
}

GList *
mrp_task_get_assigned_resources (MrpTask *task)
{
	GList         *list = NULL;
	GList         *assignments;
	GList         *node;
	MrpAssignment *assignment;
	MrpResource   *resource;
	
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);

	assignments = mrp_task_get_assignments (task);
	
	for (node = assignments; node; node = node->next) {
		assignment = MRP_ASSIGNMENT (node->data);
		resource   = MRP_RESOURCE (
			mrp_assignment_get_resource (assignment));
		
		list = g_list_prepend (list, resource);
	}

	list = g_list_sort (list, mrp_resource_compare);

	g_list_free (assignments);

	return list;
}

gint
mrp_task_compare (gconstpointer a, gconstpointer b)
{
	return strcmp (MRP_TASK (a)->priv->name, MRP_TASK (b)->priv->name);
}

void
mrp_task_reset_constraint (MrpTask *task)
{
	g_return_if_fail (MRP_IS_TASK (task));

	if (task->priv->constraint.type != MRP_CONSTRAINT_ASAP) {
		task->priv->constraint.type = MRP_CONSTRAINT_ASAP;
		g_object_notify (G_OBJECT (task), "constraint");
	}
}

/**
 * mrp_task_get_cost:
 * @task: A task object
 * 
 * 
 * 
 * Return value: the cost to complete the task
 **/
gfloat
mrp_task_get_cost (MrpTask *task)
{
	GList *resources, *l;
	gfloat total_cost = 0; /*, total_cost_overtime = 0;*/
	gfloat cost; /*, cost_overtime;*/
	
	g_return_val_if_fail (MRP_IS_TASK (task), 0);
	
	resources = mrp_task_get_assigned_resources (task);

	for (l = resources; l; l = l->next) {		
		mrp_object_get (l->data, "cost", &cost, NULL);
		/*mrp_object_get (l->data, "cost_overtime", &cost_overtime, NULL);*/
		
		total_cost += (mrp_task_get_duration (task)/3600) * cost;
		/*total_cost_overtime += (mrp_task_get_duration (task)/3600) * cost_overtime;*/
	}

	g_list_free (resources);
	
	return total_cost;
}

/* Boxed types. */

static MrpConstraint *
mrp_constraint_copy (const MrpConstraint *dst)
{
	MrpConstraint *copy;

	copy = g_new (MrpConstraint, 1);
	memcpy (copy, dst, sizeof (MrpConstraint));

	return copy;
}

GType
mrp_constraint_get_type (void)
{
	static GType our_type = 0;
  
	if (our_type == 0) {
		our_type = g_boxed_type_register_static ("MrpConstraint",
							 (GBoxedCopyFunc) mrp_constraint_copy,
							 (GBoxedFreeFunc) g_free);
	}
	
	return our_type;
}

/* Getters and setters. */

void
imrp_task_set_visited (MrpTask *task, gboolean visited)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->visited = visited;
}

gboolean
imrp_task_get_visited (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), FALSE);

	return task->priv->visited;
}

glong
imrp_task_get_start (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), -1);

	return task->priv->start;
}

void
imrp_task_set_start (MrpTask *task, glong start)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->start = start;
}

glong
imrp_task_get_finish (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), -1);

	return task->priv->finish;
}

void
imrp_task_set_finish (MrpTask *task, glong finish)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->finish = finish;
}

gint
imrp_task_get_duration (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), 0);

	return task->priv->duration;
}

void
imrp_task_set_duration (MrpTask *task, gint duration)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->duration = duration;
}

gint
imrp_task_get_work (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), 0);

	return task->priv->work;
}

void
imrp_task_set_work (MrpTask *task, gint work)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->work = work;
}

MrpConstraint
impr_task_get_constraint (MrpTask *task)
{
	MrpConstraint c = { 0 };
	
	g_return_val_if_fail (MRP_IS_TASK (task), c);

	return task->priv->constraint;
}

void
impr_task_set_constraint (MrpTask *task, MrpConstraint constraint)
{
	g_return_if_fail (MRP_IS_TASK (task));

	task->priv->constraint = constraint;
}

gint
imrp_task_get_depth (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), 0);
	
	return g_node_depth (task->priv->node);
}

GNode *
imrp_task_get_node (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	
	return task->priv->node;
}

GNode *
imrp_task_get_sorted_node (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	
	return task->priv->sorted_node;
}

GList *
imrp_task_peek_predecessors (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	
	return task->priv->predecessors;
}

GList *
imrp_task_peek_successors (MrpTask *task)
{
	g_return_val_if_fail (MRP_IS_TASK (task), NULL);
	
	return task->priv->successors;
}

/* This function calulates the work that needs to be set to reach the given
 * calendar duration (wrt calendars in use).
 */
gint
mrp_task_duration_to_work (MrpTask *task,
			   gint     calendar_duration)
{
	MrpTaskPriv    *priv;
	gint            nonworking_time;
	mrptime         t;
	mrptime         finish;
	mrptime         ival_start;
	mrptime         ival_end;
	GList          *intervals, *l;
	MrpInterval    *ival;
	MrpProject     *project;
	MrpCalendar    *calendar;
	MrpCalendarDay *day;
	gint            delta;

	g_return_val_if_fail (MRP_IS_TASK (task), 0);
	
	priv = task->priv;
	
	g_object_get (G_OBJECT (task), "project", &project, NULL);
	if (!project) {
		g_warning ("No project on task.");
		return 0;
	}
	
	calendar = mrp_project_get_calendar (project);
	
	nonworking_time = 0;
	finish = priv->start + calendar_duration;
	
	t = mrp_time_align_day (priv->start);
	
	while (t < finish) {
		day = mrp_calendar_get_day (calendar, t);

		intervals = mrp_calendar_get_intervals (day);
		for (l = intervals; l; l = l->next) { 
			ival = l->data;

			mrp_interval_get_absolute (ival,
						   t,
						   &ival_start,
						   &ival_end);
			
			/* Skip any intervals that are before the task. */
			if (ival_end <= priv->start) {
				continue;
			}

			if (ival_start > finish) {
				break;
			}
			
			if (mrp_interval_get_type (ival) == MRP_INTERVAL_TYPE_NONWORK) {
				delta = MIN (finish, ival_end) - MAX (priv->start, ival_start);
				nonworking_time += delta;
				
/*				g_print ("Nonwork ival: %s - %s, delta: %.2f\n",
				mrp_time_format ("%a %e %b, %k:%M", ival_start),
				mrp_time_format ("%a %e %b, %k:%M", ival_end),
				delta/(60.0*60));
*/
			}
		}
		
		t += 60*60*24;
	}

	return calendar_duration - nonworking_time;
}

#if 0
gint
mrp_task_calculate_calendar_duration (MrpTask *task,
				      gint     duration)
{
	MrpTaskPriv     *priv;
	MrpProject      *project;
	mrptime          t;
	mrptime          t1;
	mrptime          t2;
	mrptime          ival_start;
	mrptime          ival_end;
	GList           *intervals, *l;
	gint             work, w, delta;
	MrpCalendar     *calendar;
	MrpCalendarDay  *day;
	MrpInterval     *ival;
	MrpIntervalType  type;

	g_return_val_if_fail (MRP_IS_TASK (task), 0);
	
	priv = task->priv;
	
	g_object_get (G_OBJECT (task), "project", &project, NULL);
	calendar = mrp_project_get_calendar (project);
	
	t1 = priv->start;

	/* Duration is the work + any nonworking time. */
	work = priv->work;
	duration = work;
	
	w = 0;
	t = mrp_time_align_day (t1);
	
	while (w < work) {
		day = mrp_calendar_get_day (calendar, t);
		intervals = mrp_calendar_get_intervals (day);
		for (l = intervals; l; l = l->next) { 
			ival = l->data;

			mrp_interval_get_absolute (ival,
						   t,
						   &ival_start,
						   &ival_end);
			
			/* Skip any intervals that are before the task. */
			if (ival_end < t1) {
				continue;
			}

			type = mrp_interval_get_type (ival);
			
			/* If the task starts during nonworking time, delay it
			 * until after the interval.
			 */
			if (type == MRP_INTERVAL_TYPE_NONWORK &&
			    ival_start <= t1 && ival_end > t1) {
				t1 = ival_end;
				
				/* Loop again, in case there is another
				 * nonworking interval right after this one.
				 */
				continue;
			}
				
			/* Clamp the start time so that we don't add work before
			 * the task starts. Note that we don't need to clamp the
			 * end time because it will only be too high during
			 * working time, so the duration won't be affected.
			 */
			delta = ival_end - MAX (ival_start, t1);
			if (type == MRP_INTERVAL_TYPE_WORK) {
				w += delta;
/*				g_print ("[%p] Work ival   : %s - %s, delta: %.2f\n",
				task,
				mrp_time_format ("%a %e %b, %k:%M", ival_start),
				mrp_time_format ("%a %e %b, %k:%M", ival_end),
				delta/(60.0*60));
*/				
			} else {
				duration += delta;
/*				g_print ("[%p] Nonwork ival: %s - %s, delta: %.2f\n",
				task,
				mrp_time_format ("%a %e %b, %k:%M", ival_start),
				mrp_time_format ("%a %e %b, %k:%M", ival_end),
				delta/(60.0*60));
*/			
			}

			if (w >= work) {
				break;
			}
		}

		t += 60*60*24;
	}

	/*duration = imrp_task_get_duration (task);*/

	/*g_print ("Duration: %.2f\n", duration/(60.0*60));*/

	return duration;
}
#endif
